use crate::game::*;
use rand::seq::IteratorRandom;
use rayon::prelude::*;
use std::cmp::Ordering;
use std::mem::swap;

pub trait Bottable<P: GamePlayer, M: GameMove>: Game<P, M> {
    fn random_move(&self) -> Option<M> {
        let mut rng = rand::thread_rng();
        return self.moves().choose(&mut rng);
    }
    fn best_move(&self) -> Option<M> {
        let p = self.current_player();
        let moves = self.moves().collect::<Vec<M>>();
        if moves.is_empty() {
            return None;
        } else if moves.len() == 1 {
            return moves.into_iter().next();
        }
        let mut score: Vec<f64> = vec![0.0; moves.len()];
        let mut weights: Vec<f64> = vec![0.0; moves.len()];
        let mut weights_old: Vec<f64> = vec![0.0; moves.len()];
        loop {
            (&moves).into_par_iter().zip(&mut score).for_each(|(m, c)| {
                *c += (0..2048)
                    .into_par_iter()
                    .map(|_| self.clone())
                    .map(|mut game| {
                        let mut m = *m;
                        loop {
                            match game.apply(m).unwrap() {
                                MoveOutcome::Win(p2) => break if p2 == p { 1.0 } else { -1.0 },
                                MoveOutcome::Tie => break 0.0,
                                MoveOutcome::Pass => (),
                            }
                            m = game.random_move().unwrap();
                        }
                    })
                    .sum::<f64>();
            });
            let min_score = score.iter().copied().reduce(f64::min).unwrap();
            (&mut weights)
                .into_iter()
                .zip(&score)
                .for_each(|(c, s)| *c = s - min_score);
            let total: f64 = weights.iter().copied().sum::<f64>();
            // normalize scores
            if total > 0.0 {
                (&mut weights).into_iter().for_each(|c| *c /= total);
            } else {
                (&mut weights).into_iter().for_each(|c| *c = 0.0);
            }
            let deviation: f64 = weights
                .iter()
                .copied()
                .zip(weights_old.iter().copied())
                .max_by(|(a, _), (b, _)| a.partial_cmp(b).unwrap_or(Ordering::Equal))
                .map(|(s, s_old)| f64::abs(s - s_old))
                .unwrap();
            if deviation < (1.0 / f64::powf(2.0, 12.0)) {
                break Some(
                    moves
                        .into_iter()
                        .zip(weights)
                        .max_by(|(_, a), (_, b)| a.partial_cmp(b).unwrap_or(Ordering::Equal))
                        .unwrap()
                        .0,
                );
            }
            swap(&mut weights, &mut weights_old);
        }
    }
}

impl<P: GamePlayer, M: GameMove, T: Game<P, M>> Bottable<P, M> for T {}
