use crate::game::*;
use std::fmt;

#[derive(Clone)]
pub struct Four<P> {
    player: P,
    data: [(u8, [Option<P>; 6]); 7],
    filled: u8,
    done: bool,
}

impl<P: GamePlayer + fmt::Display> fmt::Display for Four<P> {
    fn fmt(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
        formatter.write_str("\n")?;
        for y in (0..6).rev() {
            formatter.write_str(" ")?;
            for x in 0..7 {
                match self.data[x].1[y] {
                    Some(p) => formatter.write_fmt(format_args!("{} ", &p)),
                    None => formatter.write_fmt(format_args!(". ")),
                }?;
            }
            formatter.write_str("\n")?;
        }
        formatter.write_str(" ")?;
        for x in 0..7 {
            formatter.write_fmt(format_args!("{} ", x + 1))?;
        }
        formatter.write_str("\n")?;
        Ok(())
    }
}

#[derive(Copy, Clone)]
pub struct FourMove(u8);
impl GameMove for FourMove {}
impl fmt::Display for FourMove {
    fn fmt(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
        (self.0 + 1).fmt(formatter)
    }
}
impl TryFrom<u8> for FourMove {
    type Error = ();
    fn try_from(num: u8) -> Result<Self, ()> {
        if num >= 1 && num <= 7 {
            Ok(Self(num - 1))
        } else {
            Err(())
        }
    }
}

pub struct FourMoves<'a, P: 'a> {
    cols: std::iter::Enumerate<std::slice::Iter<'a, (u8, [Option<P>; 6])>>,
    size: usize,
}

impl<'a, P: GamePlayer> Iterator for FourMoves<'a, P> {
    type Item = FourMove;
    fn next(&mut self) -> Option<Self::Item> {
        let r = self.cols.find(|(_, d)| d.0 < 6).map(|(x, _)| {
            self.size -= 1;
            FourMove(x as u8)
        });
        r
    }
    fn size_hint(&self) -> (usize, Option<usize>) {
        (self.size, Some(self.size))
    }
}

impl<P: GamePlayer> Four<P> {
    const DIRS: [[[i8; 2]; 2]; 4] = [
        [[0, -1], [0, 1]],
        [[-1, 0], [1, 0]],
        [[-1, -1], [1, 1]],
        [[1, -1], [-1, 1]],
    ];
    pub fn new() -> Self {
        Self {
            player: P::first(),
            data: [(0, [None; 6]); 7],
            filled: 0,
            done: false,
        }
    }
}

impl<P: GamePlayer> Game<P, FourMove> for Four<P> {
    type Moves<'a> = FourMoves<'a, P> where P: 'a;
    fn current_player(&self) -> P {
        self.player
    }
    fn moves<'a>(&'a self) -> Self::Moves<'a> {
        FourMoves {
            cols: self.data.iter().enumerate(),
            size: self.data.iter().filter(|d| d.0 < 6).count(),
        }
    }
    fn apply(&mut self, m: FourMove) -> Option<MoveOutcome<P>> {
        use MoveOutcome::*;
        if self.done {
            return None;
        }
        let x = m.0;
        if x >= 7 {
            return None;
        }
        let col = &mut self.data[x as usize];
        if col.0 >= 6 {
            return None;
        }
        let y = col.0;
        col.1[y as usize] = Some(self.player);
        col.0 += 1;
        self.filled += 1;
        for axis in Self::DIRS.iter() {
            let mut count = 1;
            for dir in axis.iter() {
                let (mut x, mut y) = (x as i8, y as i8);
                loop {
                    (x, y) = (x + dir[0], y + dir[1]);
                    if x < 0
                        || x >= 7
                        || y < 0
                        || y >= 6
                        || self.data[x as usize].1[y as usize].map_or(true, |p| p != self.player)
                    {
                        break;
                    }
                    count += 1;
                    if count >= 4 {
                        self.done = true;
                        return Some(Win(self.player));
                    }
                }
            }
        }
        self.player = self.player.next();
        Some(if self.filled >= 6 * 7 {
            self.done = true;
            Tie
        } else {
            Pass
        })
    }
}
